import {EventEmitter} from './events';
const P = Promise;

/**
 * Component is a configurable object that emits events.
 */
class Component extends EventEmitter {

  /**
   * @param void
   * @return {Object}
   */
  defaults () {
    return {};
  }

  /**
   * @constructor
   * @param {Object} options - instance configuration
   */
  constructor (options={}) {
    super();
    this.options = this.defaults();
    this.configure(options);
  }

  /**
   * @param {String} key
   * @param {Object} opts
   * @param mixed opts.fallback
   * @param {Boolean} opts.pull
   * @return mixed
   */
  option (key, opts={}) {
    let {fallback, pull} = {fallback: undefined, pull: false, ...opts};
    let {options} = this;
    let value = fallback;

    if (!options.hasOwnProperty(key) || options[key] === undefined) {
      if (fallback === undefined) {
        throw new Error(`option "${key}" is undefined`);
      }
    }
    else {
      value = options[key];
    }

    if (pull) {
      delete options[key];
    }

    return value;
  }

  /**
   * @param {Object} options
   * @return void
   */
  configure (options={}) {
    this.options = {...this.options, ...options};
  }

  /**
   * Return a Promise that wraps an event listener. Promise outcomes:
   *  - resolve with event arguments if the event was raised 
   *  - reject if the operation times out
   *
   * @async Promise
   * @param {String} event
   * @param {Number} ttl - optional time-to-live in millis; omit to wait indefinitely
   * @return {Array} - the arguments emitted with the event
   */
  await (event, ttl=null) {
    return new P((resolve, reject) => {
      let timeout;

      let listener = (...args) => {
        clearTimeout(timeout);
        resolve(args);
      };

      let expire = () => {
        this.off(event, listener);
        reject(new Error('await() operation timed out'));
      };

      if (ttl !== null) {
        timeout = setTimeout(expire, ttl);
      }

      this.once(event, listener);
    });
  }

}

/**
 * Container is a key-value store that has:
 *  - an asynchronous set() method
 *  - a synchronous get() method
 */
class Container extends Component {

  /**
   * @inheritdoc
   */
  defaults () {
    return {
      ...super.defaults(),
      bindings: new Map(),
    };
  }

  /**
   * @inheritdoc
   */
  constructor (options={}) {
    super(options);
    this.bindings = this.option('bindings', {pull: true});
  }

  /**
   * Bind a value to the container on the supplied key. Value may be any of the
   * following:
   *  - any primitive data type
   *  - a promise
   *  - a function that returns anything, including a promise
   *
   * A promised value will bind to the container when the promise settles.
   *
   * @async Promise
   * @param {String} key
   * @param mixed value
   * @param {Boolean} force
   * @return void
   */
  set (key, value, force=false) {
    let {bindings} = this;

    if (bindings.has(key) && !force) {
      throw new Error(`bind conflict on key "${key}"`);
    }

    let bind = (...args) => {
      if (!args.length) {
        throw new Error('expected an argument');
      }
      bindings.set(key, args[0]);
      return P.resolve();
    };

    if (value && typeof value === 'function') {
      value = value();
    }

    if (value && typeof value.then === 'function') {
      return value.then(bind);
    }

    return bind(value);
  }

  /**
   * @param {String} key
   * @para
   *
   */
  get (key) {
    let {bindings} = this;

    if (!bindings.has(key)) {
      throw new Error(`binding not found on key "${key}"`);
    }

    return bindings.get(key);
  }

}

export default {
  Component,
  Container,
};
