import {Component, Container} from '../../src/components';
import {EventEmitter} from 'events';
const P = Promise;

describe('components', () => {
  describe('Component', () => {


    it('is an EventEmitter', () => {
      expect(new Component()).to.be.an.instanceof(EventEmitter);
    });

    it('accepts options through the constructor', () => {
      let component = new Component({foo: 1, bar: 'two'});
      let {foo, bar} = component.options;
      expect(foo).to.equal(1);
      expect(bar).to.equal('two');
    });

    describe('#option()', () => {
      let component = new Component({foo: 1, bar: 'two', baz: undefined});

      it('returns an option value', () => {
        expect(component.option('foo')).to.equal(1);
      });

      it('complains if an option isn\'t defined', () => {
        let fn = () => component.option('non-existent-option');
        expect(fn).to.throw;
      });

      it('complains if an option is actually equal to undefined', () => {
        let fn = () => component.option('baz');
        expect(fn).to.throw;
      });

      it('optionally returns a fallback value', () => {
        expect(component.option('non-existent', {fallback: 2})).to.equal(2);
      });

      it('optionally pulls the option off the options object', () => {
        expect(component.option('bar', {pull: true})).to.equal('two');
        expect(component.options.bar).to.be.undefined; 
      });
    });

    describe('#off()', () => {
      let component = new Component();

      it('removes event listener at arity 2', () => {
        let listener = spy();
        component.on('event', listener);
        component.emit('event');
        expect(listener.calledOnce).to.be.ok;

        component.off('event', listener);
        component.emit('event');
        expect(listener.calledOnce).to.be.ok;
      });

      it('removes all listeners of a type at arity 1', () => {
        let listener1 = spy();
        let listener2 = spy();

        component.on('event', listener1);
        component.on('event', listener2);
        component.emit('event');
        expect(listener1.calledOnce).to.be.ok;
        expect(listener2.calledOnce).to.be.ok;

        component.off('event');
        component.emit('event');
        expect(listener1.calledOnce).to.be.ok;
        expect(listener2.calledOnce).to.be.ok;
      });

      it('removes all listeners at arity 0', () => {
        let listener1 = spy();
        let listener2 = spy();

        component.on('event1', listener1);
        component.on('event2', listener2);
        component.off();
        component.emit('event1');
        component.emit('event2');

        expect(listener1.called).not.to.be.ok;
        expect(listener2.called).not.to.be.ok;
      });

    });

    describe('#await()', () => {
      let component = new Component();
      let promise = component.await('event');

      it('returns a promise', () => {
        expect(promise.then).to.be.a('function');
      });

      it('the promise resolves when the event is emitted', (done) => {
        promise.then(() => done());
        component.emit('event');
      });

    });

  }); // Component

  describe('Container', () => {

    it('is a Component', () => {
      expect(new Container).to.be.an.instanceof(Component);
    });

    describe('#set()', () => {

      it('works with a promised value', () => {
        let container = new Container();
        let promise = P.resolve('value');

        return container.set('key', promise)
        .then(() => expect(container.get('key')).to.equal('value'));
      });

      it('works with a function that returns a promised value', () => {
        let container = new Container();
        let fn = () => P.resolve('value');

        return container.set('key', fn)
        .then(() => expect(container.get('key')).to.equal('value'));
      });

      it('works with a function that returns anything else', () => {
        let container = new Container();
        let fn = () => 'value';

        return container.set('key', fn)
        .then(() => expect(container.get('key')).to.equal('value'));
      });


      it('binds a string as-is', () => {
        let container = new Container();

        return container.set('key', 'value')
        .then(() => expect(container.get('key')).to.equal('value'));
      });

    });

  });

});

